/*  Copyright (C) 1995, 1996, 1998 by Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* medit.c */

#include <stdlib.h>
#include <string.h>
#include <termio.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>

/* Color is determined by the bits. for 87654321, 321 = the main color codes
 * (0 to 7). 4 = the hilited bit (1m). 765 = the background color codes
 * (0 to 7). 8 = the blinking bit (5m). normal, black background, no inverse or
 * hilite, and white text = integer code 7. */

extern char *unparse_textcolor();
extern int input_file();
extern int read_map();
extern int status_file();
extern int write_map();
extern void bigedit();
extern void color_palette();
extern void display_editblock();
extern void fix_tty();
extern void init_block();
extern void init_help();
extern void load_config();
extern void load_map();
extern void menter();
extern void print_sbar();
extern void redraw_screen();
extern void remove_sbar();
extern void reset_block();
extern void reset_tty();
extern void save_map();
extern void setup_edit();
extern void view_help();

struct helptext {
  char *text;
  struct helptext *prev;
  struct helptext *next;
};

struct helptext *helpfile;

struct textblock {
  unsigned char text[51];
  unsigned char color[50];
};

struct textblock *nulltext;
struct textblock *defaulttext;

struct screenblock {
  struct textblock *text;
  struct screenblock *xnext, *xprev, *ynext, *yprev;
};

struct screenblock *origblock;
struct screenblock *nullblock;
struct screenblock *curblock;
struct screenblock *wrapblock;

struct savefile {
  struct textblock *text;
  struct savefile *next;
};

struct savefile *savelist;

char *filename="";
char sbar[512];

int curx, cury, scurx, scury, mapx, mapy;
int mapxb, mapxt, mapyb, mapyt;
int curxprev, curyprev;

int editflag=0;
int scrollflag=0;
int modifyflag=0;
int quitsave=0;
int rangeflag=0;
int jumpflag=0;
int autosave=0;
int textcolor=7;
int helpmode=0;
int readonly=0;
int grabconfig=0;
int nrows=24;
int ncols=80;
int autocount=0;
int regenedit=1;

int intrchar;
int quitchar;

#define GRDL_VER "v1.4"
#define MAP_VERSION 1

int main(int argc, char *argv[])
{
  char *ptr, *cmd=*argv;
  int a;

  /* alter lines and columns if available */
#ifdef TIOCGWINSZ
  {
    struct winsize ws;

    if(ioctl(1, TIOCGWINSZ, &ws) != -1) {
      if(ws.ws_row > 22 && ws.ws_row < 251)
        nrows = ws.ws_row;
      if(ws.ws_col > 79 && ws.ws_col < 501)
        ncols = ws.ws_col;
    }
  }
#endif

  ptr=getenv("LINES");
  if(ptr && (a=atoi(ptr)))
    if(a > 22 && a < 251)
      nrows=a;
  ptr=getenv("COLUMNS");
  if(ptr && (a=atoi(ptr)))
    if(a > 79 && a < 501)
      ncols=a;

  load_config(cmd, "", 1);

  /* grab any flags on the command line */
  while(argc > 1) {
    --argc; ++argv;
    if(grabconfig == 1) {
      load_config(cmd, *argv, 0);
      grabconfig=2;
      continue;
    }
    if(**argv == '-') {
      for(a=1;(*argv)[a];a++) {
        switch((*argv)[a]) {
        case 'a':
          autosave=1;
          break;
        case 'c':
          grabconfig=1;
          break;
        case 'h':
          helpmode=1;
          break;
        case 'r':
          readonly=1;
          break;
        case 'v':
          fprintf(stdout, "%s: GRDL Version Reference %s\n", cmd, GRDL_VER);
          fflush(stdout);
          exit(0);
        default:
          fprintf(stderr, "%s: illegal option -- %c\n", cmd, (*argv)[a]);
          fprintf(stderr, "Usage: %s [-ahrv] [-c file] file\n", cmd);
          fflush(stderr);
          exit(1);
        }
      }
      continue;
    }
    filename=*argv;
  }

  if(!grabconfig)
    load_config(cmd, "", 0);

  /* turn off echoing */
  fix_tty();

  /* draw in default screen blocks */
  init_help();
  init_block();

  /* load map */
  a=0;
  if(*filename)
    a=read_map(1);
  if(!a) {
    status_file(0);
    redraw_screen();
  }

  /* go into main editor */
  bigedit();

  /* clean up and reset terminal */
  printf("\e[m\e[r\e[%dH\e[?25h\n", nrows);
  reset_tty();
  exit(0);
}

int status_file(int typ)
{
  struct stat statbuf;
  int rslt=0, rstr=1;
  char buf[30];

  rslt=stat(filename, &statbuf);
  strcpy(sbar, " ");
  if(*filename) {
    strcat(sbar, "File: ");
    strncat(sbar, filename, 200);
    if(typ == 2) {
      strcat(sbar, " [cannot write]");
    } else if(typ == 4) {
      strcat(sbar, " [unknown map type]");
    } else if(rslt == -1) {
      if(typ == 3) {
        strcat(sbar, " [no such file]");
        rstr=0;
      } else {
        strcat(sbar, " [new]");
      }
    } else if((statbuf.st_mode & S_IFMT) != S_IFLNK &&
              (statbuf.st_mode & S_IFMT) != S_IFREG) {
      strcat(sbar, " [directory]");
      rstr=0;
    } else if(access(filename, R_OK) == -1) {
      strcat(sbar, " [permission denied]");
      rstr=0;
    } else if(typ != 3 && typ != 5 && (readonly||access(filename, W_OK) == -1)) {
      strcat(sbar, " [read only]");
      rstr=0;
    } else if(!typ || typ == 5) {
      sprintf(buf, " %ld bytes", (long)statbuf.st_size);
      strcat(sbar, buf);
    } else if(modifyflag)
      strcat(sbar, " [no save since last change]");
  } else {
    strcat(sbar, "No File.");
    rstr=0;
  }
  sbar[ncols-1]='\0';
  strcat(sbar, " ");
  return rstr;
}

void init_block()
{
  int a;

  nulltext=malloc(sizeof(struct textblock));
  strncpy(nulltext->text,
         "                                                  ", 50);
  for(a=0;a<50;a++)
    nulltext->color[a] = 7;
  nulltext->text[50] = 0;
  defaulttext=malloc(sizeof(struct textblock));
  strncpy(defaulttext->text,
         "::::::::::::::::::::::::::::::::::::::::::::::::::", 50);
  strncpy(defaulttext->color,
         "nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn", 50);
  defaulttext->text[50] = 1;

  origblock=malloc(sizeof(struct screenblock));
  origblock->text = defaulttext;
  origblock->xprev=NULL; origblock->xnext=NULL;
  origblock->yprev=NULL; origblock->ynext=NULL;
  nullblock=malloc(sizeof(struct screenblock));
  nullblock->text = nulltext;

  curblock=origblock;

  wrapblock=malloc(sizeof(struct screenblock));
  wrapblock->text = defaulttext;

  /* initialize map pointers */
  mapxb=0; mapxt=0; mapyb=0; mapyt=0;
  mapx=-4; mapy=-2;
  curx=41; cury=11;
  curxprev=0; curyprev=0;
}

void load_config(char *cmd, char *nfile, int typ)
{
  FILE *f;
  char buf[555];
  int nline=0;

  if(!typ) {
    strncpy(buf, getenv("HOME"), 500);
    buf[500]='\0';
    strcat(buf, "/.maprc");
  } else
    strcpy(buf, "/usr/local/lib/medit/global.maprc");

  if(*nfile)
    f=fopen(nfile, "r");
  else
    f=fopen(buf, "r");

  if(!f) {
    if(typ == 1) {
      strncpy(buf, getenv("HOME"), 200);
      buf[200]='\0';
      strcat(buf, "/lib/medit/global.maprc");
      f=fopen(buf, "r");
      if(!f)
        return;
    } else {
      if(!grabconfig)
        return;
      fprintf(stderr, "%s: Cannot load config file '%s'.\n", cmd, nfile);
      fflush(stderr);
      exit(1);
    }
  }

  while(fgets(buf, 511, f)) {
    nline++;
    buf[strlen(buf)-1] = '\0';
    if(!*buf || *buf == '#')
      continue;

    if(!strncasecmp(buf, "columns", 7)) {
      if(atoi(buf+7) < 80 || atoi(buf+7) > 500) {
        fprintf(stderr, "%s: %s[%d]: Columns must be within 80 to 500 characters.\n", cmd, *nfile?nfile:".maprc", nline);
        exit(1);
      }
      ncols=atoi(buf+7);
      continue;
    } if(!strncasecmp(buf, "lines", 5)) {
      if(atoi(buf+5) < 23 || atoi(buf+5) > 250) {
        fprintf(stderr, "%s: %s[%d]: Lines must be within 23 to 250 characters.\n", cmd, *nfile?nfile:".maprc", nline);
        exit(1);
      }
      nrows=atoi(buf+5);
      continue;
    } if(!strncasecmp(buf, "color bg", 8)) {
      if(atoi(buf+8) < 0 || atoi(buf+8) > 7) {
       fprintf(stderr, "%s: %s[%d]: Background colors must be within 0 to 7.\n",
               cmd, *nfile?nfile:".maprc", nline);
       exit(1);
      }
      textcolor &= 143;
      if(atoi(buf+8)&1) textcolor += 16;
      if(atoi(buf+8)&2) textcolor += 32;
      if(atoi(buf+8)&4) textcolor += 64;
      continue;
    } if(!strncasecmp(buf, "color fg", 8)) {
      if(atoi(buf+8) < 0 || atoi(buf+8) > 15) {
        fprintf(stderr, "%s: %s[%d]: Foreground colors must be within 0 to 15.\n", cmd, *nfile?nfile:".maprc", nline);
        exit(1);
      }
      textcolor &= 240;
      textcolor += atoi(buf+8);
      continue;
    } if(!strncasecmp(buf, "color", 5)) {
      fprintf(stderr, "%s: %s[%d]: Select 'bg' or 'fg' mode.\n", cmd,
              *nfile?nfile:".maprc", nline);
      exit(1);
    } if(!strcasecmp(buf, "helpmode")) {
      helpmode=1;
      continue;
    } if(!strcasecmp(buf, "scroll off")) {
      scrollflag=0;
      continue;
    } if(!strcasecmp(buf, "scroll on")) {
      scrollflag=1;
      continue;
    } if(!strcasecmp(buf, "scroll expand")) {
      scrollflag=2;
      continue;
    } if(!strcasecmp(buf, "scroll fill")) {
      scrollflag=3;
      continue;
    } if(!strncasecmp(buf, "scroll", 6)) {
      fprintf(stderr, "%s: %s[%d]: Select 'off', 'on', 'expand', or 'fill' mode.\n",
              cmd, *nfile?nfile:".maprc", nline);
      exit(1);
    } if(!strcasecmp(buf, "textmode blink")) {
      textcolor |= 128;
      continue;
    } if(!strncasecmp(buf, "textmode", 8)) {
      textcolor &= 127;
      continue;
    } if(!strcasecmp(buf, "readonly")) {
      readonly=1;
      continue;
    } if(!strcasecmp(buf, "autosave off")) {
      autosave=0;
      continue;
    } if(!strcasecmp(buf, "autosave on")) {
      autosave=1;
      continue;
    } if(!strncasecmp(buf, "autosave", 8)) {
      fprintf(stderr, "%s: %s[%d]: Select 'off' or 'on'.\n", cmd,
              *nfile?nfile:".maprc", nline);
      exit(1);
    }
    fprintf(stderr, "%s: %s[%d]: Syntax error.\n", cmd, *nfile?nfile:".maprc",
            nline);
    exit(1);
  }
}

void init_help()
{
  struct helptext *ptr;
  char buf[512];
  FILE *f;

  helpfile=malloc(sizeof(struct helptext));
  helpfile->prev=NULL;
  helpfile->next=NULL;
  if((f=fopen("/usr/local/lib/medit/medit.help", "r")) == NULL) {
    strncpy(buf, getenv("HOME"), 200);
    buf[200]='\0';
    strcat(buf, "/lib/medit/medit.help");
    if((f=fopen(buf, "r")) == NULL) {
      helpfile->text = "Helpfiles currently not available.";
      return;
    }
  }
  ptr=helpfile;
  while(fgets(buf, 511, f)) {
    buf[strlen(buf)-1]=0;
    ptr->text=malloc(strlen(buf)+1);
    strcpy(ptr->text, buf);
    ptr->next=malloc(sizeof(struct helptext));
    ptr->next->prev=ptr;
    ptr=ptr->next;
    ptr->text=NULL;
  }
  fclose(f);
  if(helpmode)
    view_help();
}


struct screenblock *findblock(int xcrd, int ycrd)
{
  int a;

  /* if the block requested is out of range */
  if(xcrd < mapxb || xcrd > mapxt || ycrd < mapyb || ycrd > mapyt) 
    return ((scrollflag==3 && !editflag)?wrapblock:nullblock);

  /* search for new block of text */
  a=curxprev;
  while(a != xcrd) {
    if(a > xcrd) {
      curblock = curblock->xprev; a--;
    } else if(a < xcrd) {
      curblock = curblock->xnext; a++;
    }
  }
  curxprev=a;
  a=curyprev;
  while(a != ycrd) {
    if(a > ycrd) {
      curblock = curblock->yprev; a--;
    } else if(a < ycrd) {
      curblock = curblock->ynext; a++;
    }
  }
  curyprev=a;

  /* return the actual screenblock */
  return curblock;
}

void erase_map()
{
  struct screenblock *tmp, *tnext;
  int a;

  tmp=findblock(mapxb, mapyb);
  while(1) {
    a=0;
    while(!a) {
      if(tmp->xnext)
        tnext=tmp->xnext;
      else {
        tnext=tmp->ynext;
        a=1;
      }
      if(!tmp->text->text[50])
        free(tmp->text);
      free(tmp);
      tmp=tnext;
    }
    if(!tmp)
      break;
    a=0;
    while(!a) {
      if(tmp->xprev)
        tnext=tmp->xprev;
      else {
        tnext=tmp->ynext;
        a=1;
      }
      if(!tmp->text->text[50])
        free(tmp->text);
      free(tmp);
      tmp=tnext;
    }
    if(!tmp)
      break;
  }
  free(nulltext);
  free(defaulttext);
  free(nullblock);
  init_block();
  modifyflag=0;
}

void print_sbar()
{
  printf("\e[%dH\e[0;1;7;37;40m%s\e[m", nrows, sbar);
}

void redraw_screen()
{
  struct textblock *curtext=NULL;
  int prev=7, color=7, text;
  int a, b;

  /* clear screen */
  printf("\e[m\e[?25h\e[H\e[2J");

  for(a=0;a<nrows;a++) {
    b=0;
    if(a == nrows-1) {
      /* print status bar */
      b=strlen(sbar);
      printf("\e[0;1;7;37;40m%s\e[m", sbar);
      prev=7; color=7;
    }
    /* print line of text */
    while(b<ncols) {
      if(!(b%10) || (a == nrows-1 && b == strlen(sbar)))
        curtext=findblock(mapx+(b/10),mapy+(a/5))->text;
      color=curtext->color[((a%5)*10)+(b%10)];
      text=curtext->text[((a%5)*10)+(b%10)];
      if(text == ' ') {
        color|=15;
        curtext->color[((a%5)*10)+(b%10)]=color;
      }
      printf("%s%c", unparse_textcolor(prev, color), text);
      ++b;
      prev=color;
    }
  }

  if(editflag)
    display_editblock();
  else {
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(prev, textcolor));
    fflush(stdout);
  }
}

void remove_sbar(int stpos)
{
  struct textblock *curtext=NULL;
  int prev=7, color=7, text;
  int a=nrows-1, b;

  rangeflag--;
  if(rangeflag < 0)
    rangeflag=0;
  if(stpos < 1)
    stpos=1;
  b=stpos-1;
  if(strlen(sbar) < stpos)
    return;
  /* print last line on screen */
  printf("\e[%d;%dH\e[m", nrows, stpos);
  while(b<strlen(sbar) && b<ncols) {
    if(!(b%10) || b == stpos-1)
      curtext=findblock(mapx+(b/10),mapy+(a/5))->text;
    color=curtext->color[((a%5)*10)+(b%10)];
    text=curtext->text[((a%5)*10)+(b%10)];
    printf("%s%c", unparse_textcolor(prev, color), text);
    ++b;
    prev=color;
  }

  /* restore cursor position */
  if(curx > 0)
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(prev, textcolor));
}

void move_screen_down(int rowl)
{
  struct textblock *curtext=NULL;
  int prev=7, color=7, text;
  int a, b;

  /* move to bottom of screen */
  if(rowl == nrows && strlen(sbar)) {
    rangeflag++;
    remove_sbar(1);
  }
  printf("\e[m\e[%dH\n", rowl);

  for(a=rowl;a<rowl+5;a++) {
    b=0;
    if(a == nrows+4 || a == nrows-1) {
      /* print status bar */
      b=strlen(sbar);
      printf("\e[0;1;7;37;40m%s\e[m", sbar);
      prev=7; color=7;
    }
    /* print line of text */
    while(b<ncols) {
      if(!(b%10) || ((a == nrows+4 || a == nrows-1) && b == strlen(sbar)))
        curtext=findblock(mapx+(b/10),mapy+(a/5))->text;
      color=curtext->color[((a%5)*10)+(b%10)];
      text=curtext->text[((a%5)*10)+(b%10)];
      if(text == ' ') {
        color|=15;
        curtext->color[((a%5)*10)+(b%10)]=color;
      }
      printf("%s%c", unparse_textcolor(prev, color), text);
      ++b;
      prev=color;
    }
  }

  /* restore cursor position */
  if(rowl == nrows) {
    cury-=5; mapy++;
  }
}

void move_screen_up()
{
  struct textblock *curtext=NULL;
  int prev=7, color=7, text;
  int a, b;

  /* clear colors */
  printf("\e[m");
  cury+=5; mapy--;

  for(a=4;a>=0;a--) {
    /* move to top of screen and scroll down */
    printf("\e[H\eM");
    b=0;
    /* print line of text */
    while(b<ncols) {
      if(!(b%10))
        curtext=findblock(mapx+(b/10),mapy+(a/5))->text;
      color=curtext->color[((a%5)*10)+(b%10)];
      text=curtext->text[((a%5)*10)+(b%10)];
      if(text == ' ') {
        color|=15;
        curtext->color[((a%5)*10)+(b%10)]=color;
      }
      printf("%s%c", unparse_textcolor(prev, color), text);
      ++b;
      prev=color;
    }
  }
}

void move_screen_right(int rowl)
{
  struct textblock *curtext=NULL;
  int prev=7, color=7, text;
  int a, b;

  /* clear colors */
  printf("\e[m");
  if(rowl == ncols-10) {
    curx-=10; mapx++;
  }

  for(a=0;a<nrows;a++) {
    /* move to top-left of screen and delete 10 chars from the right */
    if(rowl == ncols-10) {
      printf("\e[%dH\e[10P\e[%dC", a+1, ncols-10);
      b=ncols-10;
    } else {
      printf("\e[%d;%dH", a+1, rowl+1);
      b=rowl;
    }
    /* print line of text */
    while(b<rowl+10) {
      if(!(b%10) || b == rowl)
        curtext=findblock(mapx+(b/10),mapy+(a/5))->text;
      color=curtext->color[((a%5)*10)+(b%10)];
      text=curtext->text[((a%5)*10)+(b%10)];
      if(text == ' ') {
        color|=15;
        curtext->color[((a%5)*10)+(b%10)]=color;
      }
      printf("%s%c", unparse_textcolor(prev, color), text);
      ++b;
      prev=color;
    }
  }

  /* restore cursor position */
  if(rowl == ncols-10 || rowl <= strlen(sbar))
    print_sbar();
}

void move_screen_left()
{
  struct textblock *curtext=NULL;
  int prev=7, color=7, text;
  int a, b;

  /* clear colors */
  rangeflag++;
  remove_sbar(strlen(sbar)-10);
  printf("\e[m");
  curx+=10; mapx--;

  for(a=0;a<nrows;a++) {
    /* move to top-left of screen and insert 10 chars from the right */
    printf("\e[%dH\e[10@", a+1);
    b=0;
    /* print line of text */
    while(b<10) {
      if(!b)
        curtext=findblock(mapx+(b/10),mapy+(a/5))->text;
      color=curtext->color[((a%5)*10)+(b%10)];
      text=curtext->text[((a%5)*10)+(b%10)];
      if(text == ' ') {
        color|=15;
        curtext->color[((a%5)*10)+(b%10)]=color;
      }
      printf("%s%c", unparse_textcolor(prev, color), text);
      ++b;
      prev=color;
    }
  }
}

void page_up(int univ)
{
  int a;

  for(a=0;a<univ;a++) {
    if(editflag || mapy+((cury-1)/5) <= mapyb) {
      if(!a) {
        printf("\a");
        return;
      }
      break;
    }
    cury-=5;
    move_screen_up();
  }
  print_sbar();
  printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
}

void page_down(int univ)
{
  int a;

  for(a=0;a<univ;a++) {
    if(editflag || mapy+((cury-1)/5) >= mapyt) {
      if(!a) {
        printf("\a");
        return;
      }
      break;
    }
    cury+=5;
    move_screen_down(nrows);
  }
  printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
}

void cursor_up(int univ)
{
  struct screenblock *orig, *old, *new, *omain;
  int a, b=0;

  if(univ == -1)
    b=1;
  for(a=0;a<univ || b;a++) {
    if((editflag && cury%5 == 1) || (!editflag && (((scrollflag < 1 || scrollflag > 2 || b) &&
       mapy+((cury+49998)/5)-10000 < mapyb) || (scrollflag == 2 &&
       mapy+((cury+49998)/5)-10000 < mapyb-1)))) {
      if(!a) {
        if(jumpflag) {
          ++jumpflag;
          return;
        }
        printf("\a");
        return;
      }
      break;
    }
    cury--;
  }

  printf("\e[");
  if(a > 1)
    printf("%d", a);
  printf("A");
  if(cury < -4) {
    a=16-cury;
    mapy-=a/5;
    cury+=(a/5)*5;
    redraw_screen();
  } else if(cury < 1) {
    move_screen_up();
    move_screen_up();
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  } else if(cury < 6) {
    move_screen_up();
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }
  if(scrollflag == 2 && mapy+((cury-1)/5) == mapyb-1) {
    orig=findblock(mapx+((curx-1)/10),mapyb);
    if(orig == nullblock)
      return;
    modifyflag=1;
    regenedit=1;
    omain=NULL;
    for(old=orig;old;old=old->xprev) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->yprev=new;
      new->ynext=old; new->yprev=NULL;
      new->xnext=omain; new->xprev=NULL;
      if(omain)
        omain->xprev=new;
      omain=new;
    }
    omain=orig->yprev;
    for(old=orig->xnext;old;old=old->xnext) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->yprev=new;
      new->yprev=NULL; new->ynext=old;
      new->xnext=NULL; new->xprev=omain;
      if(omain)
        omain->xnext=new;
      omain=new;
    }
    mapyb--;
    move_screen_down(((cury-1)/5)*5);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }
}

void cursor_down(int univ)
{
  struct screenblock *orig, *old, *new, *omain;
  int a, b=0;

  if(univ == -1)
    b=1;
  for(a=0;a<univ || b;a++) {
    if((editflag && cury%5 == 0) || (!editflag && (((scrollflag < 1 || scrollflag > 2 || b) &&
       mapy+(cury/5) > mapyt) || (scrollflag == 2 &&
       mapy+(cury/5) > mapyt+1)))) {
      if(!a) {
        if (jumpflag) {
          ++jumpflag;
          return;
        }
        printf("\a");
        return;
      }
      break;
    }
    cury++;
  }

  printf("\e[");
  if(a > 1)
    printf("%d", a);
  printf("B");
  if(cury > (((nrows-1)/5)+2)*5) {
    a=(cury-(nrows-15))/5;
    mapy+=a;
    cury-=a*5;
    redraw_screen();
  } else if(cury > (((nrows-1)/5)+1)*5) {
    move_screen_down(nrows);
    move_screen_down(nrows);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  } else if(cury > ((nrows-1)/5)*5) {
    move_screen_down(nrows);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }

  if(scrollflag == 2 && mapy+((cury-1)/5) == mapyt+1) {
    orig=findblock(mapx+((curx-1)/10),mapyt);
    if(orig == nullblock)
      return;
    modifyflag=1;
    regenedit=1;
    omain=NULL;
    for(old=orig;old;old=old->xprev) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->ynext=new;
      new->ynext=NULL; new->yprev=old;
      new->xnext=omain; new->xprev=NULL;
      if(omain)
        omain->xprev=new;
      omain=new;
    }
    omain=orig->ynext;
    for(old=orig->xnext;old;old=old->xnext) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->ynext=new;
      new->ynext=NULL; new->yprev=old;
      new->xnext=NULL; new->xprev=omain;
      if(omain)
        omain->xnext=new;
      omain=new;
    }
    mapyt++;
    move_screen_down(((cury-1)/5)*5);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }
}

void cursor_right(int univ)
{
  struct screenblock *orig, *old, *new, *omain;
  int a, b=0;

  if(univ == -1)
    b=1;
  for(a=0;a<univ || b;a++) {
    if((editflag && curx%10 == 0) || (!editflag && (((scrollflag < 1 || scrollflag > 2 || b) &&
       mapx+(curx/10) > mapxt) || (scrollflag == 2 &&
       mapx+(curx/10) > mapxt+1)))) {
      if(!a) {
        if (jumpflag) {
          ++jumpflag;
          return;
        }
        printf("\a");
        return;
      }
      break;
    }
    curx++;
  }

  printf("\e[");
  if(a > 1)
    printf("%d", a);
  printf("C");
  if(curx > (((ncols-1)/10)+3)*10) {
    a=(curx-(ncols-40))/10;
    mapx+=a;
    curx-=a*10;
    redraw_screen();
  } else if(curx > ((ncols-1)/10)*10) {
    b=(((curx-1)/10)-(ncols-1)/10)+1;
    for(a=0;a<b;a++)
      move_screen_right(ncols-10);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }

  if(scrollflag == 2 && mapx+((curx-1)/10) == mapxt+1) {
    orig=findblock(mapxt,mapy+((cury-1)/5));
    if(orig == nullblock)
      return;
    modifyflag=1;
    regenedit=1;
    omain=NULL;
    for(old=orig;old;old=old->yprev) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->xnext=new;
      new->xnext=NULL; new->xprev=old;
      new->ynext=omain; new->yprev=NULL;
      if(omain)
        omain->yprev=new;
      omain=new;
    }
    omain=orig->xnext;
    for(old=orig->ynext;old;old=old->ynext) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->xnext=new;
      new->xnext=NULL; new->xprev=old;
      new->ynext=NULL; new->yprev=omain;
      if(omain)
        omain->ynext=new;
      omain=new;
    }
    mapxt++;
    move_screen_right(((curx-1)/10)*10);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }
}

void cursor_left(int univ)
{
  struct screenblock *orig, *old, *new, *omain;
  int a, b=0;

  if(univ == -1)
    b=1;
  for(a=0;a<univ || b;a++) {
    if((editflag && curx%10 == 1) || (!editflag && (((scrollflag < 1 || scrollflag > 2 || b) &&
       mapx+((curx+99998)/10)-10000 < mapxb) || (scrollflag == 2 &&
       mapx+((curx+99998)/10)-10000 < mapxb-1)))) {
      if(!a) {
        if (jumpflag) {
          ++jumpflag;
          return;
        }
        printf("\a");
        return;
      }
      break;
    }
    curx--;
  }

  printf("\e[");
  if(a > 1)
    printf("%d", a);
  printf("D");
  if(curx < -29) {
    a=40-curx;
    mapx-=a/10;
    curx+=(a/10)*10;
    redraw_screen();
  } else if(curx < 1) {
    b=2-(curx/10);
    for(a=0;a<b;a++)
      move_screen_left();
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  } else if(curx < 11) {
    move_screen_left();
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }

  if(scrollflag == 2 && mapx+((curx-1)/10) == mapxb-1) {
    orig=findblock(mapxb,mapy+((cury-1)/5));
    if(orig == nullblock)
      return;
    modifyflag=1;
    regenedit=1;
    omain=NULL;
    for(old=orig;old;old=old->ynext) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->xprev=new;
      new->xprev=NULL; new->xnext=old;
      new->yprev=omain; new->ynext=NULL;
      if(omain)
        omain->ynext=new;
      omain=new;
    }
    omain=orig->xprev;
    for(old=orig->yprev;old;old=old->yprev) {
      new=malloc(sizeof(struct screenblock));
      new->text = defaulttext;
      old->xprev=new;
      new->xprev=NULL; new->xnext=old;
      new->yprev=NULL; new->ynext=omain;
      if(omain)
        omain->yprev=new;
      omain=new;
    }
    mapxb--;
    move_screen_right(((curx-1)/10)*10);
    printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }
}

void cursor_home()
{
  int a=((mapy*5)+(cury-1)), b=0;

  if(editflag)
    a=((cury-1)%5);
  if(a > 0)
    cursor_up(a);
  else if(a < 0)
    cursor_down(a*(-1));
  else
    ++b;

  if(editflag)
    a=((curx-1)%10);
  else
    a=((mapx*10)+(curx-1));

  if(a > 0)
    cursor_left(a);
  else if(a < 0)
    cursor_right(a*(-1));
  else
    ++b;

  if(b == 2)
    printf("\a");
}

void set_scroll(int univ)
{
  int redo=scrollflag;

  scrollflag+=((univ-1)%4)+1;
  if(scrollflag > 3)
    scrollflag-=4;
  remove_sbar(27);
  if(!scrollflag)
    strcpy(sbar, " Cannot Scroll Beyond Map ");
  else if(scrollflag == 1)
    strcpy(sbar, " Scroll Beyond Map Without Expansion ");
  else if(scrollflag == 2)
    strcpy(sbar, " Expand Map When Scrolling Beyond Border ");
  else
    strcpy(sbar, " Fill Border With Default Block ");
  if(scrollflag > 2 || redo > 2) {
    redraw_screen();
  } else {
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  }
}

void bigedit()
{
  int a, qflag=0, univ;
  char buf[512], ch;

  /* jump into the editor until quit command is reached */
  while(!qflag) {
    if(autosave && modifyflag && *filename && ++autocount > 1000) {
      autocount=0;
      remove_sbar(14);
      strcpy(sbar, " Autosave... ");
      print_sbar();
      printf("\e[?25l");
      fflush(stdout);
      write_map();
      modifyflag=0;
    }
    if(rangeflag) {
      sprintf(buf,
         " Area: %d, TOP:%d BOT:%d LFT:%d RGT:%d, Pos: %d^%c %d^%c ",
         (unsigned int)((((mapyb*(-1))+mapyt+1)*5)*(((mapxb*(-1))+mapxt+1)*10)),
         mapyb*(-5), (mapyt*5)+4, mapxb*(-10), (mapxt*10)+9,
         abs((mapx*10)+(curx-1)), ((mapx*10)+(curx-1))>0?'E':
         ((mapx*10)+(curx-1))<0?'W':' ', abs((mapy*5)+(cury-1)),
         ((mapy*5)+(cury-1))>0?'S':((mapy*5)+(cury-1))<0?'N':' ');
      rangeflag++;
      remove_sbar(strlen(buf)+1);
      strcpy(sbar, buf);
      print_sbar();
      printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    }

    /* grab a character from input */
    fflush(stdout);
    univ=1;
    ch=getchar();
    if(ch == 'U'-64) {
      int univ_init=0;

      rangeflag++;
      remove_sbar(8);
      strcpy(sbar, " C-u 4 ");
      print_sbar();
      printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
      univ=4;
      while(isdigit(ch=getchar())) {
        if(ch == '0' && !univ_init) {
          strcpy(sbar, " C-u 0 ");
          print_sbar();
          printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
          univ=0;
          fflush(stdout);
          continue;
        }
        if(!univ_init) {
          univ=0;
          univ_init=1;
        }
        if(univ > 99999) {
          printf("\a");
          fflush(stdout);
          continue;
        }
        univ*=10;
        univ+=ch-48;
        sprintf(sbar, " C-u %d ", univ);
        print_sbar();
        printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
        fflush(stdout);
      }
      rangeflag++;
      remove_sbar(1);
      strcpy(sbar, "");
    }
    if(univ < 1 && ch != 'G'-64)
      continue;

    if(ch == 'X'-64) {
      remove_sbar(6);
      strcpy(sbar, " C-x ");
      print_sbar();
      printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
      /* grab rest of string */
      ch=getchar();
      switch(ch) {
      case 'G'-64:
        remove_sbar(7);
        strcpy(sbar, " Quit ");
        printf("\a");
        print_sbar();
        printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
        continue;
      case 's': case 'S': case 'S'-64:
      case 'w': case 'W': case 'W'-64:
        save_map();
        continue;
      case 'f': case 'F': case 'F'-64:
      case 'l': case 'L': case 'L'-64:
      case 'v': case 'V': case 'V'-64:
        load_map();
        continue;
      case 'x': case 'X': case 'X'-64:
        remove_sbar(10);
        status_file(0);
        print_sbar();
        printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
        continue;
      case 'c': case 'C': case 'C'-64:
      case 'q': case 'Q': case 'Q'-64:
        quitsave=1;
        save_map();
        if(quitsave) {
          if(quitsave == 1) {
            remove_sbar(1);
            strcpy(sbar, "");
          }
          /* quit the program */
          printf("\e[m\e[r\e[%dH\e[?25h\n", nrows);
          reset_tty();
          exit(0);
        }
        continue;
      default:
        remove_sbar(1);
        strcpy(sbar, "");
        if(ch == 27)
          break;
        printf("\a");
        continue;
      }
    }

    switch(ch) {
    /* check for escape */
    case 27:
      /* grab another character */
      while((ch=getchar()) == 27);
      a=0;
      switch(ch) {
	/* toggle autosave */
	case 'a': case 'A':
          autosave+=univ%2;
          if(autosave > 1)
            autosave=0;
          remove_sbar(21);
          sprintf(sbar, " Autosave Turned %s ", autosave?"On":"Off");
          print_sbar();
          printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
          break;
	/* switch to the color palette */
	case 'c': case 'C':
	  color_palette();
	  break;
	/* remove status bar */
	case 'd': case 'D':
	  remove_sbar(1);
	  strcpy(sbar, "");
	  break;
	/* edit default map block */
	case 'e': case 'E':
	  setup_edit();
	  break;
	/* view help screen */
	case 'h': case 'H':
	  view_help();
	  break;
	/* reset current block to default */
	case 'k': case 'K':
	  if(editflag)
	    printf("\a");
	  else
	    reset_block();
	  break;
	/* toggle scroll rendition */
	case 's': case 'S':
          if(editflag)
            printf("\a");
          else
            set_scroll(univ);
          break;
        /* view map range */
        case 'r': case 'R':
        if(editflag)
          printf("\a");
        else if(rangeflag) {
          remove_sbar(1);
          strcpy(sbar, "");
        } else
          rangeflag++;
        break;
        /* refresh screen */
        case 'l': case 'L':
          redraw_screen();
          break;
        /* cursor movement */
        case 'p': case 'P':
          cursor_up(-1);
          break;
        case 'n': case 'N':
          cursor_down(-1);
          break;
        case 'f': case 'F':
          cursor_right(univ*10);
          break;
        case 'b': case 'B': case 'I'-64:
          cursor_left(univ*10);
          break;
        /* page up */
        case 'v': case 'V':
          page_up(univ);
          break;
        /* print file status */
        case 'x': case 'X':
          remove_sbar(10);
          status_file(1);
          print_sbar();
          printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
          break;
        /* terminate control sequence */
        case 'G'-64:
          remove_sbar(7);
          strcpy(sbar, " Quit ");
          printf("\a");
          print_sbar();
          printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
          break;
        /* move to top left */
        case '<':
          jumpflag=1;
          cursor_up(-1);
          cursor_left(-1);
          if(jumpflag == 3)
            printf("\a");
          jumpflag=0;
          break;
        /* move to bottom right */
        case '>':
          jumpflag=1;
          cursor_down(-1);
          cursor_right(-1);
          if(jumpflag == 3)
            printf("\a");
          jumpflag=0;
          break;
        /* cursor keys */
        case '[':
        ch=getchar();
        switch(ch) {
        /* up */
        case 'A':
          cursor_up(univ);
          break;
        /* down */
        case 'B':
          cursor_down(univ);
          break;
        /* right */
        case 'C':
          cursor_right(univ);
          break;
        /* left */
        case 'D':
          cursor_left(univ);
          break;
        /* home */
        case 'G': case 'H':
          cursor_home();
          break;
        case '[':
          ch=getchar();
          switch(ch) {
          default:
            a=1;
            break;
          }
          break;
        default:
          if(isdigit(ch)) {
            int fch=ch-48;

            while(isdigit(ch=getchar())) {
	      fch*=10; fch+=ch-48;
            }
            if(ch == '~') {
              switch(fch) {
              case 1:
                cursor_home();
                break;
              case 2:
                cursor_right(univ);
                break;
              case 3:
                cursor_left(univ);
                break;
	      case 4:
                set_scroll(univ);
                break;
              case 5:
                page_up(univ);
                break;
              case 6:
                page_down(univ);
                break;
              case 24:
                cursor_left(univ);
                break;
              case 25:
                cursor_down(univ);
                break;
              case 28:
	        view_help();
                break;
              default:
                a=1;
              }
            }
          } else 
            a=1;
        }
        break;
        default:
          a=1;
      }
      if(a)
        printf("\a");
      continue;
    break;
    /* CR/LF character & other line controls */
    case 13: case 10:
    case 'N'-64:
      cursor_down(univ);
      continue;
    /* up character */
    case 'P'-64:
      cursor_up(univ);
      continue;
    /* backward character */
    case 'B'-64: case 'H'-64: case 127:
      cursor_left(univ);
      break;
    /* forward character */
    case 'F'-64:
      cursor_right(univ);
      break;
    /* refresh screen */
    case 'L'-64:
      redraw_screen();
      break;
    /* terminate control sequence */
    case 'G'-64:
      remove_sbar(7);
      strcpy(sbar, " Quit ");
      printf("\a");
      print_sbar();
      printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
      break;
    /* left of map */
    case 'A'-64:
      cursor_left(-1);
      break;
    /* right of map */
    case 'E'-64:
      cursor_right(-1);
      break;
    /* page down */
    case 'V'-64:
      page_down(univ);
      break;
    /* tab character */
    case 'I'-64:
      cursor_right(univ*10);
      break;
    /* shell escape */
    case '^'-64: {
      char *ptr=getenv("SHELL");

      if(!ptr) {
        remove_sbar(11);
        strcpy(sbar, " No Shell ");
        print_sbar();
        break;
      }
      reset_tty();
      printf("\e[m\e[%dH\e[?25h\nExecuting: %s\n", nrows, ptr?ptr:"/bin/sh");
      fflush(stdout);
      system(ptr);
      fix_tty();
      if(!rangeflag)
        status_file(1);
      redraw_screen();
      break;
    }
    default:
      menter(ch, univ);
    }
  }
}

void reset_tty()
{
  struct termio pty;

  fflush(stdout);
  ioctl(0, TCGETA, &pty);
  pty.c_lflag |= ICANON;
  pty.c_lflag |= ECHO;
  pty.c_iflag |= IXON;
  pty.c_iflag |= IXOFF;
  pty.c_cc[VINTR] = intrchar;
  pty.c_cc[VQUIT] = quitchar;
  ioctl(0, TCSETA, &pty);
}

void fix_tty()
{
  struct termio pty;

  fflush(stdout);
  ioctl(0, TCGETA, &pty);
  pty.c_lflag &= ~ICANON;
  pty.c_lflag &= ~ECHO;
  pty.c_iflag &= ~IXON;
  pty.c_iflag &= ~IXOFF;
  intrchar = pty.c_cc[VINTR];
  pty.c_cc[VINTR] = 0;
  quitchar = pty.c_cc[VQUIT];
  pty.c_cc[VQUIT] = 0;
  ioctl(0, TCSETA, &pty);
}

void menter(int ch, int univ)
{
  struct screenblock *new;
  struct textblock *newt;
  int a, b, c;

  if((ch >= -1 && ch < 32) || ch == 127) {
    printf("\a");
    return;
  }

  for(a=0;a<univ;a++) {
    if(editflag) {
      new=wrapblock;
    } else {
      new=findblock(mapx+((curx-1)/10),mapy+((cury-1)/5));
      if(new == nullblock || new == wrapblock) {
        printf("\a");
        return;
      }
    }
    modifyflag=1;
    if(!editflag && new->text->text[50]) {
      newt = malloc(sizeof(struct textblock));
      memcpy(newt, new->text, sizeof(struct textblock));
      newt->text[50] = 0;
      new->text = newt;
    }
    new->text->text[(((cury-1)%5)*10)+((curx-1)%10)] = ch;
    new->text->color[(((cury-1)%5)*10)+((curx-1)%10)] = textcolor&255;

    if((editflag && curx%10 == 0) || mapx+(curx/10) > mapxt) {
      printf("%c", ch);
      return;
    }
    printf("%c", ch);
    curx++;
    if(curx > ((ncols-1)/10)*10) {
      b=(((curx-1)/10)-(ncols-1)/10)+1;
      for(c=0;c<b;c++)
        move_screen_right(ncols-10);
      printf("\e[m\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    }
  }
  return;
}

void reset_block()
{
  struct screenblock *new;

  modifyflag=1;
  new=findblock(mapx+((curx-1)/10),mapy+((cury-1)/5));
  new->text = defaulttext;
  redraw_screen();
}

void color_palette()
{
  int ch, c=0, quitflag=0, sside;

  printf("\e[m\e[2J\e[2;28H\e[?25l\e[35mC \e[1;33mO \e[0;31mL \e[1;36mO \e[0;32mR   \e[33mP \e[1;31mA \e[1;34mL \e[1;35mE \e[0;36mT \e[1;32mT \e[37mE\e[m");
  printf("\e[4;13H\e[1;4m BACKGROUND COLOR \e[4;43H TEXT COLOR \e[m");
  printf("\e[6;15H\e[1;30mBlack (Default)\e[7;15H\e[31mRed\e[8;15H\e[32mGreen\e[9;15H\e[33mYellow\e[10;15H\e[34mBlue\e[11;15H\e[35mPurple\e[12;15H\e[36mCyan\e[13;15H\e[37mWhite\e[6;45H\e[30mBlack\e[7;45H\e[0;31mRed\e[8;45H\e[32mGreen\e[9;45H\e[33mBrown\e[10;45H\e[34mBlue\e[11;45H\e[35mPurple\e[12;45H\e[36mCyan\e[13;45H\e[37mGrey (Default)\e[14;45H\e[1;30mDark Gray\e[15;45H\e[31mPink\e[16;45H\e[32mLight Green\e[17;45H\e[33mYellow\e[18;45H\e[34mLight Blue\e[19;45H\e[35mLight Purple\e[20;45H\e[36mLight Cyan\e[21;45H\e[37mWhite");
  printf("\e[16;16H\e[1;4m SPECIAL EFFECTS \e[m\e[18;15H\e[33;5mText Blink Enable\e[%dH\e[0;1;31mCursor Keys to move.\e[%dH\e[32mPress RETURN when finished. Press SPACE BAR to toggle Special Effects.", nrows-1, nrows);
  if(textcolor&64) c += 4;
  if(textcolor&32) c += 2;
  if(textcolor&16) c += 1;
  printf("\e[%d;12H\e[30m>\e[37m", c+6);
  c=0;
  if(textcolor&128)
    printf("\e[18;13H*");
  printf("\e[%d;42H>", (textcolor&15)+6);
  sside=2;

  while(!quitflag) {
    fflush(stdout);
    ch=getchar();

    if((ch == 13) || (ch == 10)) {
      quitflag=1;
      continue;
    }
    if(ch == 27) {
      while((ch=getchar()) == 27);
      if(ch != '[') {
	printf("\a");
	continue;
      }
      ch=getchar();
      if(ch == 'A') {
	if(sside == 2) {
	  if((textcolor&15) == 0) continue;
	  printf(" \e[A>");
	  textcolor -= 1;
        } else {
	  if((textcolor&112) == 0) continue;
	  printf(" \e[A>");
	  textcolor -= 16;
        }
	continue;
      }
      if(ch == 'B') {
	if(sside == 2) {
	  if((textcolor&15) == 15) continue;
	  printf(" \e[B>");
	  textcolor += 1;
        } else {
	  if((textcolor&112) == 112) continue;
	  printf(" \e[B>");
	  textcolor += 16;
        }
	continue;
      }
      if(ch == 'C') {
	if(sside == 1) {
	  sside = 2;
	  printf("\e[30m>\e[%d;42H\e[37m>", (textcolor&15)+6);
        }
	continue;
      }
      if(ch == 'D') {
	if(sside == 2) {
	  sside = 1;
	  if(textcolor&64) c += 4;
	  if(textcolor&32) c += 2;
          if(textcolor&16) c += 1;
	  printf("\e[30m>\e[%d;12H\e[37m>", c+6);
	  c=0;
	}
	continue;
      }
    }

    if(ch == 32) {
      textcolor += 128;
      textcolor &= 255;
      printf("\e[18;13H%s", textcolor&128?"*":" ");
      if(sside == 2) {
	printf("\e[%d;42H", (textcolor&15)+6);
      } else {
        if(textcolor&64) c += 4;
        if(textcolor&32) c += 2;
        if(textcolor&16) c += 1;
        printf("\e[%d;12H", c+6);
	c=0;
      }
      continue;
    }
    printf("\a");
  }
  redraw_screen();
}

char *unparse_textcolor(int prev, int color)
{
  static char buf[20];
  char buff[20];
  int c=0;

  if(prev == color)
    return "";
  strcpy(buff, "");
  if((prev&128) != (color&128)) {
    if(!(prev&128))
      strcat(buff, "5;");
    else
      strcat(buff, "25;");
  }
  if((prev&8) != (color&8)) {
    if(!(prev&8))
      strcat(buff, "1;");
    else
      strcat(buff, "22;");
  }
  if((prev&7) != (color&7)) {
    sprintf(buf, "3%d;", color&7);
    strcat(buff, buf);
  }
  if((prev&112) != (color&112)) {
    if(color&64) c += 4;
    if(color&32) c += 2;
    if(color&16) c += 1;
    sprintf(buf, "4%d;", c);
    strcat(buff, buf);
  }
  if((c=strlen(buff)))
    buff[c-1]='m';
  sprintf(buf, "\e[%s", buff);
  return (buf);
}

void view_help()
{
  struct helptext *topptr, *botptr;
  int a, ch, quitflag=0, scrl;

  printf("\e[m\e[2J\e[?25l\e[1;33m\e[HMap Editor Helptext\e[1;30HUse Cursor Keys to Scroll\e[1;64HType 'Q' to exit.\n\e[31m");
  for(a=0;a<ncols;a++)
    printf("-");
  printf("\e[%dH", nrows);
  for(a=0;a<ncols;a++)
    printf("-");
  printf("\e[0;32m\e[3;%dr\e[3H", nrows-1);
  topptr=helpfile;
  botptr=helpfile;
  for(a=0;a<nrows-3;a++) {
    printf("%s%s", botptr->text, a==nrows-4?"":"\n");
    if(!botptr->next)
      break;
    if(a != nrows-4)
      botptr=botptr->next;
  }

  while(!quitflag) {
    fflush(stdout);
    ch=getchar();

    scrl=0;
    switch(ch) {
    case 27:
      while((ch=getchar()) == 27);
      if(ch == '[') {
        ch=getchar();
        if(ch == 'A' || ch == 'D')
          scrl=1;
        else if(ch == 'B' || ch == 'C')
          scrl=2;
        else if(ch == '5') {
          getchar();
          scrl=1;
        } else if(ch == '6') {
          getchar();
          scrl=2;
        } else {
          printf("\a");
          continue;
        }
        break;
      }
    case 'h': case 'H': case 'k': case 'K': case '-':
      scrl=1;
      break;
    case 'l': case 'L': case 'j': case 'J': case '+':
      scrl=2;
      break;
    case 'q': case 'Q': case 'x': case 'X':
      quitflag=1;
      break;
    default:
      printf("\a");
      continue;
    }
    if(scrl == 1) {
      if(!topptr->prev) {
        printf("\a");
        continue;
      }
      topptr=topptr->prev;
      botptr=botptr->prev;
      printf("\e[3H\eM%s\n", topptr->text);
    } else if(scrl == 2) {
      if(!botptr->next || !botptr->next->text) {
        printf("\a");
        continue;
      }
      topptr=topptr->next;
      botptr=botptr->next;
      printf("\e[%dH\n%s\e[3H\n", nrows-1, botptr->text);
    }
  }
  if(helpmode) {
    /* quit the program */
    printf("\e[m\e[r\e[%dH\e[?25h\n", nrows);
    reset_tty();
    exit(0);
  }
  printf("\e[r");
  redraw_screen();
}

/* file commands */
void save_map()
{
  int ch, quitflag=0, rslt=0;
  char buf[512];

  if(quitsave && !modifyflag)
    return;
  if(!modifyflag || quitsave) {
    remove_sbar(66);
    if(!modifyflag)
      strcpy(sbar, " Modification was not made to the map. Still wish to save? [y/n] ");
    else
      strcpy(sbar, " Modified buffers exist. Do you wish to save? [y/n] ");
    print_sbar();
    printf("\e[?25l\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    while(!quitflag) {
      fflush(stdout);
      ch=getchar();
      switch(ch) {
      case 'G'-64:
        remove_sbar(7);
        strcpy(sbar, " Quit ");
        printf("\a");
        print_sbar();
        printf("\e[?25h\e[%d;%dH%s", cury, curx,
               unparse_textcolor(7, textcolor));
        quitsave=0;
        return;
      case 'n': case 'N':
        printf("\e[?25h");
        remove_sbar(1);
        strcpy(sbar, "");
        return;
      case 'y': case 'Y':
        quitflag=1;
        break;
      default:
        printf("\a");
      }
    }
  }

  while(1) {
   if(*filename) {
    remove_sbar(20);
    strcpy(buf, filename);
    buf[61]='\0';
    sprintf(sbar, " Save as '%s'? [y/n] ", buf);
    print_sbar();
    printf("\e[?25l\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    quitflag=0;
    while(!quitflag) {
      fflush(stdout);
      ch=getchar();
      switch(ch) {
      case 'G'-64:
        remove_sbar(7);
        strcpy(sbar, " Quit ");
        printf("\a");
        print_sbar();
        printf("\e[?25h\e[%d;%dH%s", cury, curx,
               unparse_textcolor(7, textcolor));
        quitsave=0;
        return;
      case 'n': case 'N':
        printf("\e[?25h");
        rslt=0;
        quitflag=1;
        break;
      case 'y': case 'Y':
        rslt=1;
        quitflag=1;
        break;
      default:
        printf("\a");
      }
    }
    if(rslt) {
      remove_sbar(13);
      strcpy(sbar, " Writing... ");
      print_sbar();
      fflush(stdout);
      write_map();
      if(quitsave)
        quitsave=2;
      modifyflag=0;
      return;
    }
    }
    rslt=input_file(buf, 0);
    if(rslt == -1) {
      quitsave=0;
      return;
    } else if(!rslt) {
      remove_sbar(1);
      strcpy(sbar, "");
      printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
      return;
    }
    filename=malloc(rslt+1);
    strcpy(filename, buf);
  }
}

int input_file(char *buf, int typ)
{
  int ch, a;

  printf("\e[%dH\e[0;1;7;37;40m", nrows);
  for(a=0;a<ncols;a++)
    printf(" ");
  printf("\e[%dH %s: Enter Filename: \e[?25h", nrows, typ?"Load":"Save");
  a=0;
  while(1) {
    fflush(stdout);
    ch=getchar();
    if((ch == 'H'-64 || ch == 127) && a) {
      printf(" ");
      a--;
      continue;
    } else if((ch > 32 && ch < 127) && a < ncols-24) {
      printf("%c", ch);
      buf[a]=ch;
      a++;
      continue;
    } else if(ch == 'G'-64) {
      sprintf(sbar, "%500s", "");
      remove_sbar(7);
      strcpy(sbar, " Quit ");
      printf("\a");
      print_sbar();
      printf("\e[?25h\e[%d;%dH%s", cury, curx,
             unparse_textcolor(7, textcolor));
      return -1;
    } else if(ch == 10 || ch == 13)
      break;
    printf("\a");
  }
  buf[a]='\0';
  sprintf(sbar, "%500s", "");
  return a;
}

void write_number(FILE *f, int num, int len)
{
  unsigned int a=num;
  int b;

  for(b=len-1;b>=0;b--)
    fputc((a >> b*8)&255, f);
}

int read_number(FILE *f, int len)
{
  unsigned int a=0, b, c;

  for(b=0;b<len;b++) {
    a<<=8;
    c=fgetc(f);
    if(c == EOF)
      return -1;
    a+=c;
  }
  return a;
}

int save_block(struct textblock *text)
{
  struct savefile *ptr, *last=NULL, *new;

  for(ptr=savelist;ptr;ptr=ptr->next) {
    last=ptr;
    if(!memcmp(ptr->text->text, text->text, 50) &&
       !memcmp(ptr->text->color, text->color, 50))
      return 0;
  }

  new=malloc(sizeof(struct savefile));
  new->next=NULL;
  new->text=text;
  if(!last)
    savelist=new;
  else
    last->next=new;
  return 1;
}

void write_blocks(FILE *f)
{
  struct savefile *ptr;

  for(ptr=savelist;ptr;ptr=ptr->next) {
    fwrite(ptr->text->text, 50, 1, f);
    fwrite(ptr->text->color, 50, 1, f);
  }
}

void free_blocks()
{
  struct savefile *ptr, *pnext;

  for(ptr=savelist;ptr;ptr=pnext) {
    pnext=ptr->next;
    free(ptr);
  }
  savelist=NULL;
}

int read_block(FILE *f, struct textblock *text)
{
  if(!fread(text->text, 50, 1, f) || !fread(text->color, 50, 1, f))
    return 0;
  text->text[50] = 1;
  return 1;
}

void write_bdef(FILE *f, struct textblock *text, int len)
{
  struct savefile *ptr;
  int a=0;

  for(ptr=savelist;ptr;ptr=ptr->next) {
    if(!memcmp(ptr->text->text, text->text, 50) &&
       !memcmp(ptr->text->color, text->color, 50)) {
      write_number(f, a, len);
      return;
    }
    a++;
  }
  write_number(f, 0, len);
}

int write_map()
{
  FILE *f;
  int rslt=status_file(0), ntext=0, a, b, clen=1;
  char buf[512];

  printf("\e[?25h");
  if(!rslt) {
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    return 0;
  }
  strncpy(buf, filename, 500);
  buf[500]='\0';
  strcat(buf, "~");
  rename(filename, buf);
  if((f=fopen(filename, "w")) == NULL) {
    status_file(2);
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    return 0;
  }
  fprintf(f, "MEDIT");
  write_number(f, MAP_VERSION, 1);
  write_number(f, mapyb, 4);
  write_number(f, mapxb, 4);
  write_number(f, (mapxb*(-1))+mapxt+1, 4);
  savelist=NULL;
  save_block(defaulttext);
  for(a=mapyb;a<=mapyt;a++)
    for(b=mapxb;b<=mapxt;b++)
      ntext+=save_block(findblock(b, a)->text);
  write_number(f, ntext, 4);
  write_blocks(f);
  for(a=ntext;a>255;a/=256)
    clen++;
  for(a=mapyb;a<=mapyt;a++)
    for(b=mapxb;b<=mapxt;b++)
      write_bdef(f, findblock(b, a)->text, clen);
  fclose(f);
  free_blocks();
  sprintf(sbar, "%13s", "");
  remove_sbar(10);
  status_file(0);
  print_sbar();
  printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
  return 1;
}

void load_map()
{
  char *oldfile;
  char buf[512];
  int rslt;

  quitsave=1;
  save_map();
  if(!quitsave)
    return;
  quitsave=0;

  rslt=input_file(buf, 1);
  if(rslt == -1) {
    return;
  } else if(!rslt) {
    remove_sbar(1);
    strcpy(sbar, "");
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    return;
  }

  remove_sbar(13);
  strcpy(sbar, " Reading... ");
  print_sbar();
  fflush(stdout);
  oldfile=filename;
  filename=malloc(rslt+1);
  strcpy(filename, buf);
  if(read_map(0) < 1) {
    free(filename);
    filename=oldfile;
    return;
  }
}

void reset_map()
{
  status_file(4);
  free(nulltext);
  free(defaulttext);
  free(nullblock);
  init_block();
  redraw_screen();
}

int read_map(int typ)
{
  struct textblock **textlist;
  struct screenblock **scrlist=NULL, **prevlist=NULL;
  struct stat statbuf;
  int rslt=status_file(typ?5:3), ntext=0;
  int a, b, c, d=0, clen=1, qflag=0, maplen, mapver;
  long fsize;
  char buf[5];
  FILE *f;

  if(!rslt) {
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    return 0;
  }
  if((f=fopen(filename, "r")) == NULL) {
    status_file(1);
    print_sbar();
    printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(7, textcolor));
    return 0;
  }
  strcpy(sbar, "");
  erase_map();
  free(origblock);

  /* check map header */
  fread(buf, 5, 1, f);
  if(strncmp(buf, "MEDIT", 5)) {
    fclose(f);
    reset_map();
    return -1;
  }

  /* check map version */
  if((mapver=read_number(f, 1)) != 1) {
    fclose(f);
    reset_map();
    return -1;
  }

  /* read in header arguments */
  mapyb=read_number(f, 4);
  mapxb=read_number(f, 4);
  maplen=read_number(f, 4);
  mapxt=(mapxb+maplen)-1;
  ntext=read_number(f, 4);
  for(a=ntext;a>255;a/=256)
    clen++;

  /* check that filesize = header arguments */
  stat(filename, &statbuf);
  fsize=statbuf.st_size-((((long)ntext+1L) * 100L)+22L);
  if(fsize < 1L || (fsize/clen) % (long)maplen) {
    fclose(f);
    reset_map();
    return -1;
  }

  if(!read_block(f, defaulttext)) {
    fclose(f);
    reset_map();
    return -1;
  }
  defaulttext->text[50] = 1;

  textlist = (struct textblock **)malloc(sizeof(struct textblock *)*
                                  (ntext+1));
  for(a=0;a<ntext;a++) {
    textlist[a] = (struct textblock *)malloc(sizeof(struct textblock));
    if(!read_block(f, textlist[a])) {
      fclose(f);
      for(b=0;b<a+1;b++)
        free(textlist[b]);
      free(textlist);
      reset_map();
      return -1;
    }
  }

  b=mapyb;
  while(!qflag) {
    scrlist=(struct screenblock **)malloc(sizeof(struct screenblock *)*maplen);
    for(a=0;a<maplen;a++)
      scrlist[a] = (struct screenblock *)malloc(sizeof(struct screenblock));
    for(a=0;a<maplen;a++) {
      c=read_number(f, clen);
      if(c == -1) {
        if(!a && b != mapyb && b >= 0) {
          qflag=1;
          break;
        }
        for(a=0;a<ntext;a++)
          free(textlist[a]);
        free(textlist);
        reset_map();
        fclose(f);
        return -1;
      }
      if(c > ntext)
        c=0;
      if(!c)
        scrlist[a]->text=defaulttext;
      else
        scrlist[a]->text=textlist[c-1];
      scrlist[a]->xprev=(!a)?NULL:scrlist[a-1];
      scrlist[a]->xnext=(a == maplen-1)?NULL:scrlist[a+1];
      scrlist[a]->yprev=(!prevlist)?NULL:prevlist[a];
      scrlist[a]->ynext=NULL;
      if(prevlist)
        prevlist[a]->ynext=scrlist[a];

      if(c && c <= d)
        textlist[c-1]->text[50] = 1;
      else
        d=c;
    }
    if(!b)
      curblock=scrlist[mapxb*(-1)];
    prevlist=scrlist;
    b++;
  }
  mapyt=b-2;
  regenedit=1;

  fclose(f);
  status_file(0);
  redraw_screen();
  return 1;
}

void setup_edit()
{
  struct textblock *newt;

  if(editflag) {
    editflag=0;
    curx=scurx;
    cury=scury;
    remove_sbar(1);
    strcpy(sbar, "");
    redraw_screen();
    return;
  }
  editflag=1;
  scurx=curx; scury=cury;
  curx=41; cury=11;
  if(regenedit) {
    newt = malloc(sizeof(struct textblock));
    memcpy(newt, defaulttext, sizeof(struct textblock));
    newt->text[50] = 1;
    defaulttext = newt;
    wrapblock->text = defaulttext;
    regenedit=0;
  }
  remove_sbar(24);
  strcpy(sbar, " Editing Default Block ");
  if(scrollflag < 3) {
    print_sbar();
    display_editblock();
  } else
    redraw_screen();
}

void display_editblock()
{
  int a=0, b, prev, color, text;
  printf("\e[m\e[10;40H.----------.\e[11;40H|\e[11;51H|\e[12;40H|\e[12;51H|\e[13;40H|\e[13;51H|\e[14;40H|\e[14;51H|\e[15;40H|\e[15;51H|\e[16;40H`----------'");
  prev=7;

  while(a<5) {
    b=0;
    printf("\e[%d;41H", a+11);
    while(b<10) {
      color=defaulttext->color[((a%5)*10)+(b%10)];
      text=defaulttext->text[((a%5)*10)+(b%10)];
      printf("%s%c", unparse_textcolor(prev, color), text);
      ++b;
      prev=color;
    }
    ++a;
  }

  printf("\e[%d;%dH%s", cury, curx, unparse_textcolor(prev, textcolor));
  fflush(stdout);
}
